/*
    Copyright 2010 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <windows.h>
#include <wincrypt.h>

typedef unsigned char   u8;



#define VER     "0.1"
#define KEY     "\x52\x09\x6A\xD5\x30\x36\xA5\x38\xBF\x40\xA3\x9E\x81\xF3\xD7\xFB\x7C\xE3\x39\x82\x9B\x2F\xFF\x87\x34\x8E\x43\x44\xC4\xDE\xE9\xCB"



void ubiglpwd(u8 *msg, u8 *pwd);
int unbase64(u8 *in, int insz, u8 *out, int outsz);
void std_err(int type);



int main(int argc, char *argv[]) {
    FILE    *fd;
    u8      buff[4096],
            *p;

    setbuf(stdout, NULL);

    fputs("\n"
        "Ubisoft Game Launcher password decrypter "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\n"
            "Usage: %s <settings.ini/password>\n"
            "\n", argv[0]);
        goto quit;
    }

    fd = fopen(argv[1], "rb");
    if(fd) {
        while(fgets(buff, sizeof(buff), fd)) {
            for(p = buff; *p && (*p != '\n') && (*p != '\r'); p++);
            *p = 0;
            p = strchr(buff, '=');
            if(!p) continue;
            *p++ = 0;
            if(stristr(buff, "pass")) {
                ubiglpwd(buff, p);
            } else {
                printf("- %s: %s\n", buff, p);
            }
        }
        fclose(fd);
    } else {
        ubiglpwd("password", argv[1]);
    }

quit:
    printf("\nPress RETURN to quit\n");
    fgetc(stdin);
    return(0);
}



void ubiglpwd(u8 *msg, u8 *pwd) {
    HCRYPTPROV  hProv;
    HCRYPTHASH  hHash;
    HCRYPTKEY   hKey;
    DWORD       len;

    len = strlen(pwd);  // the output buffer is ever shorter than the input one
    len = unbase64(pwd, len, pwd, len);

    if(!CryptAcquireContext(
        &hProv,
        NULL,
        NULL,   // MS strong
        PROV_RSA_FULL,
        0)) std_err(1);

    if(!CryptCreateHash(
        hProv,
        0x8003,
        0,
        0,
        &hHash)) std_err(1);

    if(!CryptHashData(
        hHash,
        KEY,
        sizeof(KEY) - 1,
        0)) std_err(1);

    if(!CryptDeriveKey(
        hProv,
        0x6602,
        hHash,
        0,
        &hKey)) std_err(1);

    if(!CryptDecrypt(
        hKey,
        0,
        TRUE,
        0,
        pwd,
        &len)) std_err(1);

    pwd[len] = 0;
    printf("- %s: %S\n", msg, (wchar_t *)pwd); // unicode

    CryptDestroyKey(hKey);
    CryptDestroyHash(hHash);
    CryptReleaseContext(hProv, 0);
}



int unbase64(u8 *in, int insz, u8 *out, int outsz) {
    int     xlen,
            a   = 0,
            b   = 0,
            c,
            step;
    u8      *limit,
            *data,
            *p;
    static const u8 base[128] = {   // supports also the Gamespy base64 and URLs
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x3e,0x00,0x3e,0x00,0x3f,
        0x34,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,
        0x0f,0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x3e,0x00,0x3f,0x00,0x3f,
        0x00,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f,0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,
        0x29,0x2a,0x2b,0x2c,0x2d,0x2e,0x2f,0x30,0x31,0x32,0x33,0x00,0x00,0x00,0x00,0x00
    };

    if(insz < 0) insz = strlen(in);
    xlen = ((insz >> 2) * 3) + 1;    // NULL included in output for text
    if(outsz < xlen) return(-1);
    data = in;

    p = out;
    limit = data + insz;

    for(step = 0; /* data < limit */; step++) {
        do {
            if(data >= limit) {
                c = 0;
                break;
            }
            c = *data;
            data++;
            if((c == '=') || (c == '_')) {  // supports also the Gamespy base64
                c = 0;
                break;
            }
        } while(c && ((c <= ' ') || (c > 0x7f)));
        if(!c) break;

        switch(step & 3) {
            case 0: {
                a = base[c];
                break;
            }
            case 1: {
                b = base[c];
                *p++ = (a << 2)        | (b >> 4);
                break;
            }
            case 2: {
                a = base[c];
                *p++ = ((b & 15) << 4) | (a >> 2);
                break;
            }
            case 3: {
                *p++ = ((a & 3) << 6)  | base[c];
                break;
            }
        }
    }
    *p = 0;
    return(p - out);
}



void std_err(int type) {
    if(type) {
        printf("\n"
            "Error: error during the usage of the cryptography (0x%08x)\n"
            "\n", (int)GetLastError());
    } else {
        perror("\nError");
    }
    exit(1);
}


